# This file contains the source code to run the face detector on the validation and test set images
import torch
from PIL import Image
import os
from facenet_pytorch import MTCNN
from tqdm import tqdm

def detect_faces(img_files,thresholds,max_detections,logger,gpu_index=None):
  """
  This function detects faces on given img paths by using MTCNN model
  """

  # device for the inference
  device = torch.device(f"cuda:{gpu_index}") if gpu_index >= 0 else torch.device("cpu")

  # if select_largest is False, all bboxes are sorted by their detection probabilities, otherwise their size
  face_detector = MTCNN(min_face_size=40, factor=0.709, thresholds=thresholds, keep_all=True,select_largest=False,device=device)

  detections = {}

  for img_file in tqdm(img_files):
    try:
      # load image; RGB PIL Image for MTCNN
      img = Image.open(img_file)

      # get bounding boxes and confidences
      faces = face_detector.detect(img,landmarks=True)

      if faces[0] is not None:
        # they are all sorted based on their detection probability
        bboxes,qualities,landmarks = faces

      else:
        # no faces detected; it will not be saved in .csv file, because it is a empty list
        bboxes, qualities,landmarks = [], [],[]
        logger.warning("No face was found for image %s", os.path.basename(img_file))

      detections[os.path.basename(img_file)] = [(qualities[i],bboxes[i],landmarks[i]) for i in range(min(max_detections, len(qualities)))]

    except Exception as e:
      logger.error("File %s: error %s",img_file,e)

  return detections


def save_detections(detections,saving_path):
  """
  This writes all detection results to specified file.
  """

  with open(saving_path, "w") as f:

    f.write("FILE,DETECTION_SCORE,BB_X,BB_Y,BB_WIDTH,BB_HEIGHT,REYE_X,REYE_Y,LEYE_X,LEYE_Y,NOSE_X,NOSE_Y,RMOUTH_X,RMOUTH_Y,LMOUTH_X,LMOUTH_Y\n")

    for image in sorted(detections.keys()):

      for (score,bbox,lmark) in detections[image]:

        f.write("%s,%3.2f,%3.2f,%3.2f,%3.2f,%3.2f,%3.2f,%3.2f,%3.2f,%3.2f,%3.2f,%3.2f,%3.2f,%3.2f,%3.2f,%3.2f\n" % (image, score,
                                                                                      bbox[0],bbox[1],bbox[2]-bbox[0],bbox[3]-bbox[1],
                                                                                      lmark[0][0],lmark[0][1],
                                                                                      lmark[1][0],lmark[1][1],
                                                                                      lmark[2][0],lmark[2][1],
                                                                                      lmark[3][0],lmark[3][1],
                                                                                      lmark[4][0],lmark[4][1]))



# This file contains a script to run the MTCNN (baseline face detector) on the validation and test set images,
# and writes them into a file
import yamlparser
import logging
import os
import multiprocessing
from ..dataset import read_ground_truth
from ..face_detection import detect_faces,save_detections
from functools import partial

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
logger = logging.getLogger("UCCS.FaceRec")

def read_configuration_file():
  
  parent_direct = os.path.dirname(os.path.dirname(__file__))
  cfg = yamlparser.config_parser(default_config_files=[os.path.join(parent_direct, "configs/baseline_config.yaml")])

  if not os.path.exists(cfg.result_directory):
    os.mkdir(cfg.result_directory)

  return cfg


def main():

  # get command line arguments
  cfg = read_configuration_file()

  # load detection protocol
  logger.info("Loading UCCS %s detection protocol", cfg.which_set)
  data = read_ground_truth(cfg.data_directory,cfg.which_set)
  img_names = sorted(data.keys())

  # get the paths of images
  image_directory = cfg.image_directory
  img_files = [os.path.join(image_directory,img_name) for img_name in img_names]

  if cfg.detection.parallel == 0:
    logger.info("Detecting faces in %d images sequentially",len(img_names))
    detections = detect_faces(img_files,cfg.detection.thresholds,cfg.detection.max_detections,logger,cfg.gpu)

  else:
    # parallelization; split data into chunks
    logger.info("Detecting faces in %d images using %d parallel processes", len(img_names), cfg.detection.parallel)

    pool = multiprocessing.Pool(cfg.detection.parallel)

    partial_detect_faces = partial(detect_faces,thresholds=cfg.detection.thresholds,max_detections=cfg.detection.max_detections,logger=logger,gpu_index=cfg.gpu)
    # Split image names into chunks for parallel processing
    chunks = [([d for i, d in enumerate(img_files) if i % cfg.detection.parallel == p]) for p in range(cfg.detection.parallel)]

    # Perform parallel processing
    results = pool.map(partial_detect_faces, chunks)

    # Combine the results from all processes
    detections = {}
    for result in results:
      detections.update(result)

    pool.close()
    pool.join()

  result_file = cfg.detection.results
  logger.info("Writing detections to file %s", result_file)
  save_detections(detections, result_file)

if __name__ == "__main__":
    main()



# This file contains source code related to face normalization and feature extraction
import sys
import numpy as np
import torch.utils.data as data
import torch
from torchvision import transforms
import cv2
from skimage import transform
import os
import subprocess
import gdown

def download_MagFace(args,logger):

    """
    It downloads the MagFace model and its model weights unless it was downloaded before
    """

    magFace_directory = os.path.join(os.path.dirname(os.path.realpath(__file__)),"MagFace")
    weights_path = os.path.join(magFace_directory,"magface_epoch_00025.pth")

    # downloading magface repo
    if not os.path.exists(magFace_directory):
        os.mkdir(magFace_directory)

        # the baseline requires MagFace model
        repository_url = "https://github.com/IrvingMeng/MagFace.git"

        # construct the Git clone command
        git_clone_command = ["git", "clone", repository_url, magFace_directory]

        # run the Git clone command
        try:
            subprocess.run(git_clone_command, check=True)
        except subprocess.CalledProcessError as e:
            logger.error("Exception on process, rc= %s output= %s", e.returncode, e.output)
            os.rmdir(magFace_directory)
            sys.exit(1)

    # downloading magface iresnet100 model weights
    if not os.path.exists(weights_path):
        # download the default model weights from google drive (backbone--iresnet100)
        model_weights_url = "https://drive.google.com/uc?id=1Bd87admxOZvbIOAyTkGEntsEz3fyMt7H"
        
        # use gdown to download the file
        try:
            gdown.download(model_weights_url, weights_path, quiet=False)
        except Exception as e:
            logger.error(f"Error: Failed to download model weights from Google Drive - {e}")
            sys.exit(1)

    # append MagFace module
    sys.path.append(magFace_directory)
    args.recognition.unfreeze()
    args.recognition.resume = weights_path

class ImgData(data.Dataset):
    """
    It takes data containing the paths of images, bounding boxes/landmarks to align/crop the faces
    """
    def __init__(self, data,which_set=None,image_size=112, align=True, transform=None):

        self.img_paths,bboxes,landmarks = data
        self.points = landmarks if align else bboxes

        self.which_set = which_set

        if align:
            assert self.points != None
            self.align = align
            self.mode = "arcface"
            self.arcface_src = np.array(
                                [[38.2946, 51.6963], [73.5318, 51.5014], [56.0252, 71.7366],
                                [41.5493, 92.3655], [70.7299, 92.2041]],
                                dtype=np.float32)
            self.arcface_src = np.expand_dims(self.arcface_src, axis=0)


        self.transform = transform
        self.image_size = image_size

    def __getitem__(self, index):

        # get image,bboxes and landmarks
        img_path = self.img_paths[index]

        points = self.points[index]

        if self.align:
            assert points.shape == (points.shape[0],5,2) # facial landmarks
        else:
            assert points.shape == (points.shape[0],4) # bboxes

        img_name = os.path.basename(img_path)

        if not os.path.isfile(img_path):
            raise Exception('{} does not exist'.format(img_path))

        # read img
        img = cv2.imread(img_path)

        if img is None:
            raise Exception('{} is empty'.format(img_path))

        # crop and align faces in the image for MagFace
        faces = []
        for point in points:

            # align face
            if self.align:
                M, _ = self.estimate_norm(point)
                face = cv2.warpAffine(img, M, (self.image_size, self.image_size), borderValue=0.0)

            # otherwise, just crop the face
            else:
                # convert it to x1,y1,x2,y2 format
                point[:,2] += point[:,0]
                point[:,3] += point[:,1]

                # crop the face based on th bbox : x1,y1,x2,y2
                face = img[int(point[1]):int(point[3]), int(point[0]):int(point[2])]
                face = cv2.resize(face,((self.image_size, self.image_size)))

            assert face.shape == (self.image_size, self.image_size,3)

            face = self.transform(face)

            faces.append(face)

        if self.which_set == "gallery":
            return faces[0],img_name

        # stack all faces in the image for the extraction
        faces = torch.stack(faces)

        return faces,img_name

    def estimate_norm(self,lmk):
        # gets the facial landmark (reye,leye,nose,rmouth,lmouth)
        assert lmk.shape == (5, 2)
        tform = transform.SimilarityTransform()
        lmk_tran = np.insert(lmk, 2, values=np.ones(5), axis=1)
        min_M = []
        min_index = []
        min_error = float('inf')
        if self.mode == 'arcface':
            if self.image_size == 112:
                src = self.arcface_src
            else:
                src = float(self.image_size) / 112 * self.arcface_src

        for i in np.arange(src.shape[0]):
            tform.estimate(lmk, src[i])
            M = tform.params[0:2, :]
            results = np.dot(M, lmk_tran.T)
            results = results.T
            error = np.sum(np.sqrt(np.sum((results - src[i])**2, axis=1)))

            if error < min_error:
                min_error = error
                min_M = M
                min_index = i

        return min_M, min_index

    def __len__(self):
        return len(self.img_paths)


def inference_dataloader(data,which_set,batch_size_perImg,workers):
    """
    It builds the dataloader to apply preprocesssing and get the input ready for the model.
    """
    # for preprocessing transformations
    trans = transforms.Compose([
        transforms.ToTensor(),
        transforms.Normalize(
            mean=[0., 0., 0.],
            std=[1., 1., 1.]),
    ])

    if which_set=="gallery":
        # each gallery identity has 10 samples, embeddings will be saved based on the identity
        batch_size_perImg = 10 if batch_size_perImg<10 else (batch_size_perImg // 10) * 10
    else:
        # each image has different number of faces to be extracted, embeddings will be saved based on the image
        batch_size_perImg = 1

    inf_dataset = ImgData(
        data,
        which_set=which_set,
        image_size=112,
        align=True,
        transform=trans
    )

    inf_loader = torch.utils.data.DataLoader(
        inf_dataset,
        batch_size=batch_size_perImg,
        num_workers=workers,
        pin_memory=True,
        shuffle=False)

    return inf_loader,batch_size_perImg

def build_model(args):
    """
    It builds the MagFace model
    """
    from .MagFace.inference.network_inf import builder_inf

    model_args = args.recognition
    # magface requires cpu_mode argument
    model_args.cpu_mode = args.gpu < 0
    model = builder_inf(model_args)

    device = torch.device(f"cuda:{args.gpu}" if args.gpu >= 0 else "cpu")
    model = model.to(device)

#    if len(args.gpu) > 1:
#        model = torch.nn.DataParallel(model,device_ids=args.gpu)

    # switch to the evaluation mode
    model.eval()

    return model,device

def save_features_information(data,img_names,which_set,features,save_directory,batch_size_perImg):
    """
    It saves all information including detection_score,bboxes,landmarks and embedding as .pth file
    """

    if which_set == "gallery":
        # save the all information based on the each identity
        for idx in range(0,batch_size_perImg,10):

            identity = img_names[idx].split("_")[0]
            landmarks = np.stack([data[name][1][0] for name in img_names[idx:idx+10]])

            image_values = {
                "landmarks": landmarks,
                "embeddings": features[idx:idx + 10, :]
            }

            torch.save(image_values,os.path.join(save_directory,f"{identity}.pth"))

    else:
        # otherwise, save the all information of the each image
        image_values = {
            "detection_scores": data[img_names[0]][0],
            "bboxes": data[img_names[0]][1],
            "landmarks": data[img_names[0]][2],
            "embeddings": features
        }

        torch.save(image_values,os.path.join(save_directory,f"{img_names[0][:-4]}.pth"))



# This file contains functionality to read the protocol files
import pandas as pd
import os
import csv

def read_ground_truth(data_dir,which_set):
    """
    It reads the ground_truth .csv file that contains bboxes,subject_ids and/or landmarks in the images
    For the test set, it only reads the image names as keys of the dictionary; the values are ``None``.

    It returns a dictionary based on image keys and their values that contain subject id and bboxes/landmarks
    """
    csv_path = os.path.join(data_dir,which_set + ".csv")
    if not os.path.exists(csv_path):
        raise ValueError("The ground truth file %s.csv does not exist in %s (Data Directory)" % (which_set, data_dir))
    
    csv_file = pd.read_csv(csv_path)
    img_files = csv_file["FILE"].unique()

    data = {}
    
    for img in img_files:

        ground_truth = csv_file[csv_file["FILE"] == img]
        #each image can have subject_id, bboxes and/or landmarks
        subject_id = ground_truth["SUBJECT_ID"].astype(int).tolist()

        if which_set == "gallery":
            #for gallery; ground truth is facial landmarks
            ground_tr = ground_truth[["REYE_X", "REYE_Y", "LEYE_X", "LEYE_Y", "NOSE_X", "NOSE_Y",
                            "RMOUTH_X", "RMOUTH_Y", "LMOUTH_X", "LMOUTH_Y"]].values.reshape(len(subject_id),5,2)
            face_id = None
        else:
            #for validation; ground truth is bounding boxes
            ground_tr = ground_truth[["FACE_X","FACE_Y","FACE_WIDTH","FACE_HEIGHT"]].values
            face_id = ground_truth["FACE_ID"].astype(int).tolist()
 
        data[img] = (face_id,subject_id,ground_tr)

    return data

def read_detections(csv_path):
    """
    It reads the .csv file that contains bboxes and/or landmarks of the detections in the images

    It returns a dictionary based on image keys and their values that contain detection score, bounding boxes and landmarks if they exist
    """
    if not os.path.exists(csv_path):
        raise ValueError("The score file '%s' does not exist" % csv_path)

    score_file = pd.read_csv(csv_path)

    data = {}

    for img,group in score_file.groupby(by="FILE"):

        #each image can have detection scores,bboxes and landmarks
        detection_scores = group[["DETECTION_SCORE"]].values
        bboxes = group[["BB_X","BB_Y","BB_WIDTH","BB_HEIGHT"]].values

        try:
            landmarks = group[["REYE_X", "REYE_Y", "LEYE_X", "LEYE_Y", "NOSE_X", "NOSE_Y",
                            "RMOUTH_X", "RMOUTH_Y", "LMOUTH_X", "LMOUTH_Y"]].values.reshape(-1,5,2)
        except:
            landmarks = None
        
        data[img] = (detection_scores,bboxes,landmarks)

    return data

def read_recognitions(score_file):
    """
    It reads the score file
    """
    data = {}
    with open(score_file) as data_file:
        # read data line by line
        reader = csv.reader(data_file)
        # Skip the header line
        next(reader)

        for row in reader:

            if not len(row):
            # skip empty lines
                continue
            image,detection_score,bbox, cosine_scores = _get_values(row)
            
            if image not in data:
                data[image] = ([], [], [])  # Initialize with empty lists for detection scores, boxes, and similarity scores

            data[image][0].append(detection_score)
            data[image][1].append(bbox)
            data[image][2].append(cosine_scores)

    return data

def _get_values(splits):
  
  # each line should has 1006 values according to the score file structure
  if len(splits) != 1006 :
    raise ValueError("Cannot interpret score file line '%s'" % ",".join(splits))

  image = splits[0] + ".jpg"
  detection_score = float(splits[1])
  bbx = [float(v) for v in splits[2:6]] # x, y, width, height
  scores = [float(v) for v in splits[6:]] # scores in order id1_score, id2_score, ...

  return image, detection_score, bbx,scores


# This file contains functionality to process scores in order to create O-ROC plots
import matplotlib.pyplot as plt
import numpy as np
from torchvision.ops import box_iou
import torch
from termcolor import cprint

def get_iou_pairwise(ground_truth_bboxes,detection_bboxes):
    """
    It computes the standard iou scores between ground truth and detection bboxes
    """
    # convert it to tensor and x1,y1,x2,y2 format 
    gr = torch.Tensor(ground_truth_bboxes)
    gr[:,2] += gr[:,0]
    gr[:,3] += gr[:,1]
    dt = torch.Tensor(detection_bboxes)
    dt[:,2] += dt[:,0]
    dt[:,3] += dt[:,1]
    
    assert gr.shape == (len(gr),4) and dt.shape==(len(dt),4)

    # get the pairwise of iou scores between ground truth and detections
    iou_pairwise = box_iou(gr,dt)

    return iou_pairwise.numpy()

def assign_detections(ground_truth,proposals,overlap_threshold=0.5,exclude=None):
    """
    It matches bounding boxes(proposals) with ground truth (if possible) or assigns it as a misdetection.
    Propasals can come from the detection files or scoring files

    A detection can match with only one unique ground truth in this matching procedure.

    Note that gallery faces will be excluded from the results when their face_ids are given in the the parameter of 'exclude'. 
    """
    assigned_detections = {}
    # store all detection scores [scores..],matched_overlaps -> [(indice_proposal,subject_id),..], misdetections-> [indices..]
    for img, (face_ids, subject_ids, ground_truth_bboxes) in ground_truth.items():
        
        if img not in proposals:
            continue

        detection_scores,prop_bboxes,_ = proposals[img]
        prop_bboxes = np.array(prop_bboxes)

        iou_pairwise = get_iou_pairwise(ground_truth_bboxes,prop_bboxes)

        # initialize arrays to keep track of matches
        matched_indices = []  # store pairs of matched ground truth and proposal indices
        unmatched_detection_indices = list(range(len(prop_bboxes)))  # initially, all proposals are unmatched

        while True:
            # find the maximum IoU in the current IoU matrix
            max_iou = np.max(iou_pairwise)
            
            if max_iou < overlap_threshold:
                break  # exit if no IoU above the threshold is found

            # find the indices of the maximum IoU
            gt_index, proposal_index = np.unravel_index(np.argmax(iou_pairwise), iou_pairwise.shape)

            # exclude gallery faces if it is given
            if exclude and face_ids[gt_index] in exclude:
                # if it is matched, dont add the detection to false positives as well as true positives
                unmatched_detection_indices.remove(proposal_index)
                iou_pairwise[gt_index, :] = 0
                iou_pairwise[:, proposal_index] = 0
                continue

            # add the matched pair to the list -> proposal index : subject_id of the ground truth
            matched_indices.append((proposal_index,subject_ids[gt_index]))
            unmatched_detection_indices.remove(proposal_index)

            # set the IoU values for the corresponding ground truth and proposals to 0
            iou_pairwise[gt_index, :] = 0
            iou_pairwise[:, proposal_index] = 0
        
        # add the image results to the dictionary
        assigned_detections[img] = (detection_scores,matched_indices,unmatched_detection_indices)

    return assigned_detections

def compute_DR_FDPI(all_matched_detections,face_numbers,image_numbers,plot_detection_numbers=False):
    """
    It computes Detection Rate (true positive rate) and False Detection Per Image (FDPI) for plotting.
    """

    positives = []
    negatives = []

    for _, (detection_scores, matched_indices, misdetections) in all_matched_detections.items():

        # detections matched with a ground-truth is assigned as a positive score 
        pos = [detection_scores[detection_ind] for detection_ind,_ in matched_indices]
        positives.extend(pos)

        # detections that are not matched with any of ground truth 
        neg = [detection_scores[detection_ind] for detection_ind in misdetections]
        negatives.extend(neg)

    cprint(f"In total: {face_numbers} faces, {len(positives)} detected faces and {len(negatives)} false detections",
           'green', attrs=['bold', 'underline'])
    positives = np.array(positives)
    negatives = np.array(negatives)

    # thresholds
    # thresholds = np.linspace(min(negatives), max(negatives), 100)
    thresholds = np.unique(negatives)

    # detection rate and false detection per image
    DR = []
    FDPI = []

    for thr in thresholds:

        dr = np.sum(positives >= thr)

        if not plot_detection_numbers:
            # get the rate
            dr /= face_numbers

        fdpi = np.sum(negatives >= thr) / image_numbers # the number of images in the probe

        DR.append(dr)
        FDPI.append(fdpi)

    return DR,FDPI

def compute_TPIR_FPIPI(all_scores,all_matched_detections,known_numbers,image_numbers,rank=1,plot_recognition_numbers=False):
    """
    It computes True Positive Identification Rate and False Postive Identification Per Image (FPIPI) for plotting.
    TPIR = True Positives / the number of known faces in the probe
    FPIPI = False Positives / the number of probe images
    """

    # get postive and negative scores
    positives = []
    negatives = []
    
    for img, (_, matched_detections, misdetections) in all_matched_detections.items():
            
        # get the score for each detection in the image
        _, _, scores =  all_scores[img]
        similarity_score = np.array(scores)

        # get score indices based on the rank
        sort = np.argsort(similarity_score[detection_indice])[::-1][:rank]
        #count true positive identification: there are two factors to be counted as true positive identification
        #1. First, the detection bbox must be matched with ground truth
        #2. Second, the similarity score of this matching should match with the subject id based on the threshold
        for detection_indice,sub_id in matched_detections:
            
            if sub_id > 0:
                best_ids = sort + 1  # because scores are sorted based on idendity number starting 1

                if sub_id in best_ids:
                    ind_id = np.where(best_ids==sub_id)[0][0]
                    pos_score = similarity_score[detection_indice][sort][ind_id]
                    positives.append(pos_score)

            # if the detection is unknown, get the max score as a negative
            else:
                negatives.append(similarity_score[detection_indice][sort[0]])
        
        # add also background detections to the negative scores
        negatives.extend([ np.max(similarity_score[mis_ind]) for mis_ind in misdetections])

    positives = np.array(positives)
    negatives = np.array(negatives)

    # thresholds
    # thresholds = np.linspace(min(negatives), max(negatives), 100)
    thresholds = np.unique(negatives)

    # tpir and fpipi for each thr
    TPIR = []
    FPIPI = []

    for thr in thresholds:

        tp = np.sum(positives >= thr)

        # normalize by the known counter
        if not plot_recognition_numbers:
            tp /= known_numbers

        fp = np.sum(negatives>=thr) / image_numbers # the number of images in the probe

        TPIR.append(tp)
        FPIPI.append(fp)

    return TPIR,FPIPI

def plot_oroc_curve(results,labels,rank,saving_path,known_numbers,linear=False,plot_recognition_numbers=False):
    """
    This function plots O-ROC curve (TPIR-FPIPI) based on given results

    results: a list of [TPIR,FPIPI]s 
    """
    # create the figure
    figure = plt.figure(figsize=(7,4.5))
    # set the plotter function based on the linear flag
    plotter = plt.semilogx if not linear else plt.plot
    
    max_fpi = 0
    min_fpi = np.inf
    for ix, label in enumerate(labels):
        plotter(results[ix][1], results[ix][0], label=label)
        
        max_fpi = max(max_fpi,results[ix][1][0]) # the first one is the biggest because its thr is the smallest
        min_fpi = min(min_fpi,results[ix][1][-1])

    plt.grid(True, color=(0.6,0.6,0.6))
    plt.title("Rank %d O-ROC curve" % rank)

    plt.legend(loc=2 if not linear else 4, prop={'size':14})
    plt.xlabel('FPI Per Image')

    if not linear:
        plt.xlim([min_fpi,max_fpi])
    else:
        plt.xlim([0,max_fpi])

    if plot_recognition_numbers:
        plt.ylim((0, known_numbers))
        plt.ylabel('TPI')
    else:
        plt.ylim((0, 1))
        plt.ylabel('TPIR Per Face')
    
    plt.tight_layout()
    plt.savefig(saving_path)

def plot_froc_curve(results,labels,saving_path,face_numbers,linear=False,plot_detection_numbers=False):
    """
    This function plots F-ROC curve (DR-FDPI) based on given results

    results: a list of [DR,FDPI]s 
    """
    # create the figure
    figure = plt.figure(figsize=(7,4.5))
    # set the plotter function based on the linear flag
    plotter = plt.semilogx if not linear else plt.plot

    max_fd = 0
    min_fd = np.inf
    for ix, label in enumerate(labels):
        plotter(results[ix][1], results[ix][0], label=label)

        max_fd = max(max_fd,results[ix][1][0]) # the first index is the biggest because its thr was the smallest
        min_fd = min(min_fd,results[ix][1][-1])

    plt.grid(True, color=(0.6,0.6,0.6))

    plt.legend(loc=2 if not linear else 4, prop={'size':14})
    plt.xlabel('False Detection Per Image')

    if not linear:
        plt.xlim([min_fd,max_fd])
    else:
        plt.xlim([0,max_fd])

    if plot_detection_numbers:
        plt.ylim((0, face_numbers))
        plt.ylabel('Number of Detections')
    else:
        plt.ylim((0, 1))
        plt.ylabel('Detection Rate')
    
    plt.tight_layout()
    plt.savefig(saving_path)



# This file contains the source code required for running the enrollment from the extracted features
import os
import numpy as np
from torch import load
from .dataset import read_ground_truth
import warnings

def average(data_dir,embeddings_dir,set="gallery"):

    data = read_ground_truth(data_dir,set)
    subject_ids = sorted(np.unique([ f'{v[1][0]:04d}' for v in data.values()]))

    gallery_embeds = []

    for id in subject_ids:
        
        #get the the path of the id
        id_embedding_pth = os.path.join(embeddings_dir,f"{id}.pth")

        #check the galllery embeddings if exists
        if not os.path.exists(id_embedding_pth):
            warnings.warn(f"The path {id_embedding_pth} does not exist.", category=Warning)
            continue

        #read the embeddings
        id_info = load(id_embedding_pth)
        id_embeddings = id_info["embeddings"]

        #take the avarage of all embeddings belonging to that id
        id_embedding = np.mean(id_embeddings,axis=0)
            
        #otherwise, get all embeddings for that identity
        gallery_embeds.append(id_embedding)

    return subject_ids,np.stack(gallery_embeds)


# This file contains code to plot F-ROC and O-ROC file for the baseline
# It can be used for either all of them or just one of them (detection and recognition)

import logging
import yamlparser
import argparse
import os
import pickle
import numpy as np
from ..dataset import read_detections,read_ground_truth,read_recognitions
from ..evaluate import assign_detections,compute_DR_FDPI,plot_froc_curve,compute_TPIR_FPIPI,plot_oroc_curve

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
logger = logging.getLogger("FaceRec.UCCS")

def read_config_file():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--tasks", "-t",
        nargs = "+",
        choices = ["detection", "recognition"],
        default = ["detection", "recognition"],
        help = "Select the tasks that should be performed in this evaluation"
    )

    parent_direct = os.path.dirname(os.path.dirname(__file__))
    cfg = yamlparser.config_parser(parser=parser,default_config_files=[os.path.join(parent_direct, "configs/baseline_config.yaml")])

    if 'detection' in cfg.tasks:
        if cfg.eval.detection.files is None:
            raise ValueError("For the detection task, --eval.detection.files are required.")

        if cfg.eval.detection.labels is None:
            cfg.eval.detection.labels = cfg.eval.detection.files

        if len(cfg.eval.detection.labels) != len(cfg.eval.detection.files):
            raise ValueError("The number of labels (%d) and results (%d) differ" % (len(cfg.eval.detection.labels), len(cfg.eval.detection.files)))

    if 'recognition' in cfg.tasks:
        if cfg.eval.recognition.files is None:
            raise ValueError("For the identification task, --eval.recognition.files are required.")

        if cfg.eval.recognition.labels is None:
            cfg.eval.recognition.labels = cfg.eval.recognition.files

        if len(cfg.eval.recognition.labels) != len(cfg.eval.recognition.files):
            raise ValueError("The number of labels (%d) and results (%d) differ" % (len(cfg.eval.recognition.labels), len(cfg.eval.recognition.labels)))

    return cfg

def main():

    # get config arguments
    cfg = read_config_file()

    # load the evaluation protocol
    logger.info("Loading UCCS %s evaluation protocol",cfg.which_set)

    #read detections based on the task (detection/identification)
    if "detection" in cfg.tasks or "recognition" in cfg.tasks:
        # read the ground truth bounding boxes of the validation set
        logger.info("Reading UCCS %s ground-truth from the protocol",cfg.which_set)
        ground_truth = read_ground_truth(cfg.data_directory,cfg.which_set)

        face_numbers = sum([len(face_ids) for face_ids,_,_ in ground_truth.values()])
        image_numbers = len(ground_truth)

        # exclude gallery faces from the result
        if cfg.eval.exclude_gallery is not None:
            
            with open(cfg.eval.exclude_gallery, 'r') as file:
                exclude = [int(line.strip()) for line in file]

            # update the number of faces in the set
            face_numbers -= len(exclude)

    # plot f-roc for the detection results if it is given
    if  "detection" in cfg.tasks:

        detection_results = []

        for idx,detection_file in enumerate(cfg.eval.detection.files):

            logger.info("Reading detections from %s (%s)", detection_file, cfg.eval.detection.labels[idx])
            detections = read_detections(detection_file)

            matched_detections = assign_detections(ground_truth,detections,cfg.eval.iou,exclude)

            logger.info("Computing DR and FDPI for %s (%s)", detection_file, cfg.eval.detection.labels[idx])
            DR,FDPI = compute_DR_FDPI(matched_detections,face_numbers,image_numbers,cfg.eval.detection.plot_numbers)

            detection_results.append((DR,FDPI))

        # plotting
        froc_path = cfg.eval.detection.froc
        logger.info("Plotting F-ROC curve(s) to file '%s'", froc_path)
        plot_froc_curve(detection_results,cfg.eval.detection.labels,froc_path,
                                 face_numbers,cfg.eval.linear,cfg.eval.detection.plot_numbers)

    # plot o-roc for the identification results if it is given
    if "recognition" in cfg.tasks:

        known_numbers = sum([ sum(np.array(subject_ids) > 0) for _,subject_ids,_ in ground_truth.values()])
        if exclude:
            known_numbers -= len(exclude)

        recognition_results = []

        for idx, score_file in enumerate(cfg.eval.recognition.files):
            logger.info("Reading scores from %s (%s)", score_file, cfg.eval.recognition.labels[idx])
            all_scores = read_recognitions(score_file)

            matched_detections = assign_detections(ground_truth,all_scores,cfg.eval.iou,exclude)

            logger.info("Computing TPIR and FPIPI for %s (%s)", score_file, cfg.eval.recognition.labels[idx])
            TPIR,FPIPI = compute_TPIR_FPIPI(all_scores,matched_detections,known_numbers,image_numbers,
                                                    cfg.eval.recognition.rank,cfg.eval.recognition.plot_numbers)

            recognition_results.append((TPIR,FPIPI))

        # plotting
        oroc_path = cfg.eval.recognition.oroc
        logger.info("Plotting O-ROC curve(s) to file '%s'", oroc_path)
        plot_oroc_curve(recognition_results,cfg.eval.recognition.labels,cfg.eval.recognition.rank,oroc_path,
                                 known_numbers,linear=cfg.eval.linear,
                                 plot_recognition_numbers=cfg.eval.recognition.plot_numbers)

if __name__ == "__main__":
    main()



# It generates the scoring .csv file before the evaluation.
import logging
import yamlparser
import os
from ..enrollment import average
from ..scoring import create_score_file

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
logger = logging.getLogger("FaceRec.UCCS")

def read_config_file():
    
    parent_direct = os.path.dirname(os.path.dirname(__file__))
    cfg = yamlparser.config_parser(default_config_files=[os.path.join(parent_direct, "configs/baseline_config.yaml")])

    if (cfg.scoring.gallery is None or cfg.scoring.probe is None):
        raise ValueError("For the scoring task, both --scoring.gallery and --scoring.probe are required.")
    
    return cfg

def main():

    # get config params
    cfg = read_config_file()

    logger.info("Loading UCCS %s scoring protocol",cfg.which_set)

    # get gallery enrollment
    logger.info("Getting UCCS gallery enrollment (average)")
    gallery_embedd_path = cfg.scoring.gallery
    subject_ids,gallery_enroll = average(cfg.data_directory,gallery_embedd_path)

    subject_ids = ["S_"+i for i in subject_ids]

    # compute scores between enrollment and probe and write them into a file
    probe_path = cfg.scoring.probe
    scoring_path = cfg.scoring.results
    logger.info("Computing scores and writing them into %s",scoring_path)

    _ = create_score_file((subject_ids,gallery_enroll),probe_path,scoring_path,cfg.gpu)

if __name__ == "__main__":
    main()


# this file contains functionality to compute similarity scores between enrolled templates and probe features, and to handle score files
import numpy as np
import os
import torch
from torch import load

def cosine(x,y, gpu_index=None):
    """
    It calculates the pairwise cosine similarity scores using PyTorch.
    Supports GPU usage.

    args:
        x: the matrix of probes (N,512)
        y: the matrix of all gallery templates (M,512)
        gpu_index: indice of the gpu to be used, for cpu it should be given 'None'
    """
    # Determine the device (CPU or GPU)
    device = torch.device("cuda:{}".format(gpu_index) if gpu_index is not None and torch.cuda.is_available() else "cpu")

    # Normalize gallery templates once and store it in a closure
    def normalize_y_once(y):
        y = torch.tensor(y, dtype=torch.float32, device=device)
        return torch.nn.functional.normalize(y, p=2, dim=1)

    if not hasattr(cosine, 'y_norm'):
        cosine.y_norm = normalize_y_once(y)

    # Convert input array x to PyTorch tensor and move it to the specified device
    x = torch.tensor(x, dtype=torch.float32, device=device)

    # Normalize x along axis 1
    x = torch.nn.functional.normalize(x, p=2, dim=1)

    # Compute the cosine similarity using PyTorch matmul
    similarity = torch.matmul(x, cosine.y_norm.t())

    return similarity.cpu().numpy() if gpu_index is not None else similarity.numpy()

def create_score_file(enrollment,probe_path,result_file,gpu_index = None):
    """
    It writes the scores between enrollment and probe to the result file
    """
    subject_ids,gallery_enroll = enrollment

    with open(result_file, "w") as f:

        f.write("FILE,DETECTION_SCORE,BB_X,BB_Y,BB_WIDTH,BB_HEIGHT," + ",".join(subject_ids)+"\n")

        for img in os.listdir(probe_path):

            #get all information including detection scores,bboxes,embeddings in that image
            probe_infos = load(os.path.join(probe_path,img))
            probe_detection_scores = probe_infos["detection_scores"]
            probe_bboxes = probe_infos["bboxes"]
            probe_embeddings = probe_infos["embeddings"]

            #cosine similarity scores
            cos_sim = cosine(probe_embeddings,gallery_enroll,gpu_index)

            for ind in range(len(cos_sim)):
                
                # write them to file
                img_name = img[:-4]
                dt_sc = probe_detection_scores[ind][0]
                x1,y1,w,h = probe_bboxes[ind]
            
                id_scores = [s for s in cos_sim[ind]]
                line = [img_name, str(dt_sc), str(x1), str(y1), str(w), str(h)]

                line.extend(map(str, id_scores))

                f.write(",".join(line) + "\n")



# This file contains a script to take the detected faces and process them with a face recognition algorithm
# It can extract features from both gallery and given set
import logging
import yaml
import yamlparser
import argparse
import os
import torch
from ..dataset import read_detections,read_ground_truth
from ..feature_extraction import download_MagFace,ImgData,inference_dataloader,build_model,save_features_information
from tqdm import tqdm

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
logger = logging.getLogger("FaceRec.UCCS")

def read_config_file():

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--disable-gallery", "-dis",
        action="store_false",
        help = "Disable the gallery feature extraction with the given set at the same time"
    )

    parent_direct = os.path.dirname(os.path.dirname(__file__))
    baseline_config = os.path.join(parent_direct, "configs/baseline_config.yaml")

    cfg = yamlparser.config_parser(parser=parser,default_config_files=[baseline_config])
        
    if not cfg.recognition.detection_file:
        raise ValueError (f"--recognition.detection_file is required.")

    if not os.path.exists(cfg.result_directory):
        os.mkdir(cfg.result_directory)

    cfg.unfreeze()

    if cfg.disable_gallery:
        read_config_file.orig_config = yaml.safe_load(open(baseline_config,'r'))
        read_config_file.orig_config["which_set"] = "gallery"

    return cfg

def update_config(cfg,orig_config):
    # update the config for gallery extraction
    cfg.update(orig_config)
    cfg.format_self()
    cfg.unfreeze()
    return cfg

def main():

    # get command line arguments
    cfg = read_config_file()

    # load extraction protocol
    logger.info("Loading {}%s extraction protocol".format("gallery and " if cfg.disable_gallery else ""), cfg.which_set)

    # download MagFace repo and its model weights if it wasn't downloaded before
    logger.info("Downloading/Activating MagFace and its model weights")
    download_MagFace(cfg,logger)

    # get the model in eval mode
    logger.info("Loading the baseline model")
    model,device = build_model(cfg)

    # sets that will be extracted
    sets = [cfg.which_set,"gallery"] if cfg.disable_gallery else [cfg.which_set]

    for set_name in sets:

        # update the cfg if the gallery will be extracted
        if cfg.which_set != set_name:
            cfg = update_config(cfg,read_config_file.orig_config)

        # read the detections for gallery,valid or test
        if cfg.which_set =="gallery":
            data = read_ground_truth(cfg.data_directory,"gallery")
        else:
            data = read_detections(cfg.recognition.detection_file)

        landmarks = [l for _,_,l in data.values()]
        img_files = list(data.keys())

        # create the path of images, it differs based on the protocol because the gallery has sub-directories
        image_directory = cfg.image_directory
        img_paths = [os.path.join(image_directory,file.split("_")[0],file) for file in img_files] if cfg.which_set == "gallery"  else [
                    os.path.join(image_directory,file) for file in img_files]

        # get the data loader
        logger.info("Creating the %s dataloader", cfg.which_set)
        inf_loader,batch_size_perImg = inference_dataloader((img_paths,None,landmarks),
                                      cfg.which_set,cfg.recognition.batch_size_perImg,cfg.recognition.workers)

        # create result dir based on the set
        result_dir = cfg.recognition.result_dir
        os.makedirs(result_dir, exist_ok=True)

        logger.info(f"Starting the {cfg.which_set} inference process")
        with torch.no_grad():

            for input, img_names in tqdm(inf_loader):

                # when batch size = 1, to avoid extra dimension increased by dataloader
                input = input[0] if batch_size_perImg == 1 else input

                # compute features
                input = input.to(device,dtype=torch.float)
                embedding_feat = model(input)

                _feat = embedding_feat.data.cpu().numpy()

                # save all information (detection scores,bboxes,landmarks,embeddings) of that image/identity
                _ = save_features_information(data,img_names,cfg.which_set,_feat,result_dir,batch_size_perImg)

if __name__ == "__main__":
    main()



# -*- coding: UTF-8 -*-

from setuptools import setup, find_packages
from os.path import abspath,dirname,join
# The only thing we do in this file is to call the setup() function with all
# parameters that define our package.
THIS_DIR: str = abspath(dirname(__file__))

with open(join(THIS_DIR, 'README.md'), encoding='utf-8') as readme:
    long_description = readme.read()
    
setup(

    # This is the basic information about your project. Modify all this
    # information before releasing code publicly.
    name = 'challenge.uccs',
    version = "3.2.1",
    description = 'Source code for running the baseline and evaluation of the third UCCS face recognition challenge',

    url = 'https://github.com/AIML-IfI/uccs-facerec-challenge',
    license = 'BSD',
    author = 'Furkan Kasim',
    author_email = 'furkan.kasim@outlook.com',

    # If you have a better, long description of your package, place it on the
    # 'doc' directory and then hook it here
    long_description = long_description,
    long_description_content_type='text/markdown',

    # This line is required for any distutils based packaging.
    # It will find all package-data inside the 'bob' directory.
    packages = find_packages('.'),
    include_package_data = True,

    # This line defines which packages should be installed when you "install"
    # this package. All packages that are mentioned here, but are not installed
    # on the current system will be installed locally and only visible to the
    # scripts of this package. Don't worry - You won't need administrative
    # privileges when using buildout.
    
    python_requires = '>=3.10',
    install_requires = open("requirements.txt").read().splitlines(),

    # This entry defines which scripts you will have inside the 'bin' directory
    # once you install the package (or run 'bin/buildout'). The order of each
    # entry under 'console_scripts' is like this:
    #   script-name-at-bin-directory = module.at.your.library:function
    #
    # The module.at.your.library is the python file within your library, using
    # the python syntax for directories (i.e., a '.' instead of '/' or '\').
    # This syntax also omits the '.py' extension of the filename. So, a file
    # installed under 'example/foo.py' that contains a function which
    # implements the 'main()' function of particular script you want to have
    # should be referred as 'example.foo:main'.
    #
    # In this simple example we will create a single program that will print
    # the version of bob.
    entry_points = {

      # scripts should be declared using this entry:
      'console_scripts' : [
          'baseline_detection = facerec.script.baseline_detection:main',
          'baseline_recognition = facerec.script.baseline_recognition:main',
          'scoring = facerec.script.scoring:main',
          'evaluation  = facerec.script.evaluation:main',
      ],
    },

    # Classifiers are important if you plan to distribute this package through
    # PyPI. You can find the complete list of classifiers that are valid and
    # useful here (http://pypi.python.org/pypi?%3Aaction=list_classifiers).
    classifiers = [
      'Development Status :: 3 - Alpha',
      'Environment :: GPU',
      'Intended Audience :: Science/Research',
      'License :: OSI Approved :: BSD License',
      'Natural Language :: English',
      'Programming Language :: Python',
      'Topic :: Scientific/Engineering :: Image Processing',
    ],
)
